/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "backprojecttreebase.h"
#include "imagesdata.h"
#include "gridtree.h"

#include "debugnew.h"

namespace grale
{

BackProjectTreeBase::BackProjectTreeBase()
{
}

BackProjectTreeBase::~BackProjectTreeBase()
{
}

void BackProjectTreeBase::init(const std::list<ImagesDataExtended *> &images, 
		  Vector2D<double> center, double areaSize, int firstLevelSubdiv, int maxDepth, bool useWeights)
{
	std::list<ImagesDataExtended *>::const_iterator it;
	std::vector<ImagesDataExtended *> imagesVector(images.size());
	int i;

	for (it = images.begin(), i = 0 ; it != images.end() ; it++, i++)
		imagesVector[i] = *it;

	storeOriginalData(imagesVector, true, true, true);

	int numSources = images.size();
	int totalPoints = 0;
	
	for (i = 0 ; i < numSources ; i++)
		totalPoints += m_numTotalPoints[i];

	m_totalPoints = totalPoints;

	m_betas.resize(totalPoints);
	m_thetas.resize(totalPoints);
	m_subThetas.resize(totalPoints);

	m_betaPointers.resize(numSources);
	m_thetaPointers.resize(numSources);

	m_invMag.resize(totalPoints);
	for (i = 0 ; i < totalPoints ; i++)
		m_invMag[i] = 1.0f;

	m_subMag0.resize(totalPoints);
	m_subMag1.resize(totalPoints);
	m_subMag2.resize(totalPoints);

	m_invMagPointers.resize(numSources);

	Vector2D<float> *pBetaPtr = &(m_betas[0]);
	Vector2D<float> *pThetaPtr = &(m_thetas[0]);
	float *pInvMagPtr = &(m_invMag[0]);

	double maxX = 0, minX = 0, maxY = 0, minY = 0;
	int pos;
	bool extSet = false;
	
	for (pos = 0, i = 0, it = images.begin() ; it != images.end() ; it++, i++)
	{
		ImagesDataExtended *pImgData = *it;
		int n = pImgData->getNumberOfImages();
		bool gotIntens = pImgData->hasIntensities();
		
		m_betaPointers[i].resize(n);
		m_thetaPointers[i].resize(n);
		m_invMagPointers[i].resize(n);
		
		for (int j = 0 ; j < n ; j++)
		{
			int p = pImgData->getNumberOfImagePoints(j);

			m_betaPointers[i][j] = pBetaPtr;
			m_thetaPointers[i][j] = pThetaPtr;
			m_invMagPointers[i][j] = pInvMagPtr;

			for (int k = 0 ; k < p ; k++, pos++)
			{
				Vector2D<double> theta = pImgData->getImagePointPosition(j,k);

				if (!extSet)
				{
					minX = theta.getX();
					maxX = theta.getX();
					minY = theta.getY();
					maxY = theta.getY();
					extSet = true;
				}
				else
				{
					minX = MIN(minX,theta.getX());
					maxX = MAX(maxX,theta.getX());
					minY = MIN(minY,theta.getY());
					maxY = MAX(maxY,theta.getY());
				}
			}			
			pBetaPtr += p;
			pThetaPtr += p;
			pInvMagPtr += p;
		}
	}
	
	m_angularScale = MIN(maxX-minX, maxY-minY)/13.0; // use something different again

	std::vector<double> distanceFractions(totalPoints);
	std::vector<Vector2D<double> > unscaledThetas(totalPoints);

	for (pos = 0, i = 0, it = images.begin() ; it != images.end() ; it++, i++)
	{
		ImagesDataExtended *pImgData = *it;
		int n = pImgData->getNumberOfImages();
		bool gotIntens = pImgData->hasIntensities();
		double D_s = pImgData->getD_s();
		double D_ds = pImgData->getD_ds();
		double distFraction = D_ds/D_s;
		
		for (int j = 0 ; j < n ; j++)
		{
			int p = pImgData->getNumberOfImagePoints(j);

			for (int k = 0 ; k < p ; k++, pos++)
			{
				distanceFractions[pos] = distFraction;
				unscaledThetas[pos] = pImgData->getImagePointPosition(j,k);
				Vector2D<double> scaledTheta = unscaledThetas[pos]/m_angularScale;
				m_thetas[pos] = Vector2D<float>(scaledTheta.getX(), scaledTheta.getY());
			}			
		}
	}

	Vector2D<double> bottomLeft = center-Vector2D<double>(areaSize, areaSize)/2.0;

	m_deflectionRows.resize(maxDepth);
	m_magnificationRows[0].resize(maxDepth);
	m_magnificationRows[1].resize(maxDepth);
	m_magnificationRows[2].resize(maxDepth);

	m_deflectionRows[0].resize(1);
	m_magnificationRows[0][0].resize(1);
	m_magnificationRows[1][0].resize(1);
	m_magnificationRows[2][0].resize(1);

	int numSidePositions = firstLevelSubdiv;

	for (int depth = 1 ; depth < maxDepth ; depth++)
	{
		int numSquares = numSidePositions*numSidePositions;

		m_deflectionRows[depth].resize(numSquares);
		m_magnificationRows[0][depth].resize(numSquares);
		m_magnificationRows[1][depth].resize(numSquares);
		m_magnificationRows[2][depth].resize(numSquares);

		numSidePositions *= 2;
	}
	
	fillInRows(firstLevelSubdiv, 0, 0, maxDepth, center, areaSize, unscaledThetas, distanceFractions);

	m_firstLevelSubdiv = firstLevelSubdiv;
	
	m_tmpBuf.resize(totalPoints*2);
	m_tmpBuf2.resize(totalPoints*2);
	m_zeroVector.resize(totalPoints);
	m_oneVector.resize(totalPoints);

	for (int p = 0 ; p < totalPoints ; p++)
	{
		m_oneVector[p] = 1.0f;
		m_zeroVector[p] = Vector2D<float>(0,0);
	}

	m_useWeights = useWeights;
}

void BackProjectTreeBase::fillInRows(int numSubdiv, int level, int offset, int maxDepth, Vector2D<double> center, double areaSize, const std::vector<Vector2D<double> > &unscaledThetas, const std::vector<double> &distanceFractions)
{
	if (level == maxDepth)
		return;

	m_deflectionRows[level][offset].resize(m_totalPoints);
	m_magnificationRows[0][level][offset].resize(m_totalPoints);
	m_magnificationRows[1][level][offset].resize(m_totalPoints);
	m_magnificationRows[2][level][offset].resize(m_totalPoints);

	for (int p = 0 ; p < m_totalPoints ; p++)
	{
		Vector2D<double> alpha;
		Vector2D<float> alphaScaled;
		double alphaxx, alphayy, alphaxy;

		getDeflectionInformation(areaSize, unscaledThetas[p]-center, alpha, alphaxx, alphayy, alphaxy);

		alpha /= m_angularScale;
		alpha *= distanceFractions[p];
		alphaScaled = Vector2D<float>(alpha.getX(), alpha.getY());

		m_deflectionRows[level][offset][p] = alphaScaled;
		m_magnificationRows[0][level][offset][p] = (float)(alphaxx*distanceFractions[p]);
		m_magnificationRows[1][level][offset][p] = (float)(alphayy*distanceFractions[p]);
		m_magnificationRows[2][level][offset][p] = (float)(alphaxy*distanceFractions[p]);
	}

	int offsetMultiplier = 4; // Since for the root node offset == 0, this can always be 4
	double subAreaSize = areaSize/(double)numSubdiv;

	for (int y = 0 ; y < numSubdiv ; y++)
	{
		for (int x = 0 ; x < numSubdiv ; x++)
		{
			double xPos = subAreaSize/2.0 + center.getX() - areaSize/2.0 + ((double)x)*subAreaSize;
			double yPos = subAreaSize/2.0 + center.getY() - areaSize/2.0 + ((double)y)*subAreaSize;

			fillInRows(2, level+1, offset*offsetMultiplier+(y*numSubdiv)+x, maxDepth, Vector2D<double>(xPos, yPos), subAreaSize, unscaledThetas, distanceFractions);
		}
	}
/*
	FillInRows(level+1, offset*4+0, maxDepth, center+Vector2D<double>(-areaSize,-areaSize)/4.0, areaSize/2.0, unscaledThetas, distanceFractions);
	FillInRows(level+1, offset*4+1, maxDepth, center+Vector2D<double>(areaSize,-areaSize)/4.0, areaSize/2.0, unscaledThetas, distanceFractions);
	FillInRows(level+1, offset*4+2, maxDepth, center+Vector2D<double>(-areaSize,areaSize)/4.0, areaSize/2.0, unscaledThetas, distanceFractions);
	FillInRows(level+1, offset*4+3, maxDepth, center+Vector2D<double>(areaSize,areaSize)/4.0, areaSize/2.0, unscaledThetas, distanceFractions);
*/
}

void BackProjectTreeBase::calculateDeflection(const GridNodeBase *pRootNode, bool intensities)
{
	CopyVector((float *)&(m_subThetas[0]), (const float *)&(m_zeroVector[0]), m_totalPoints*2);
	CopyVector((float *)&(m_subMag0[0]), (const float *)&(m_zeroVector[0]), m_totalPoints);
	CopyVector((float *)&(m_subMag1[0]), (const float *)&(m_zeroVector[0]), m_totalPoints);
	CopyVector((float *)&(m_subMag2[0]), (const float *)&(m_zeroVector[0]), m_totalPoints);

	processNode(m_firstLevelSubdiv, 0, 0, pRootNode, intensities);

	CopyVector((float *)&(m_tmpBuf[0]), (const float *)&(m_subMag0[0]), m_totalPoints);
	AddVector((float *)&(m_subMag0[0]), (const float *)&(m_subMag1[0]), m_totalPoints);
	MultiplyVector((float *)&(m_subMag1[0]), (const float *)&(m_tmpBuf[0]), m_totalPoints);
	SquareVector((float *)&(m_subMag2[0]), m_totalPoints);
	SubVector((float *)&(m_subMag1[0]), (const float *)&(m_subMag2[0]), m_totalPoints);
}

void BackProjectTreeBase::processNode(int numSubdiv, int level, int offset, const GridNodeBase *pNode, bool intensities)
{
	if (pNode->getNodeType() == GridNodeBase::TreeNode)
	{
		const GridTreeNode *pTreeNode = (const GridTreeNode *)pNode;

		int offsetMultiplier = 4; // Since for the root node offset == 0, this can always be 4
		int numChildren = numSubdiv*numSubdiv;

		for (int i = 0 ; i < numChildren ; i++)
			processNode(2, level+1, offset*offsetMultiplier+i, pTreeNode->getChild(i), intensities);
	
		/*
		ProcessNode(level+1, offset*4+0, pTreeNode->getChild(0), intensities);
		ProcessNode(level+1, offset*4+1, pTreeNode->getChild(1), intensities);
		ProcessNode(level+1, offset*4+2, pTreeNode->getChild(2), intensities);
		ProcessNode(level+1, offset*4+3, pTreeNode->getChild(3), intensities);
		*/
	}
	else // weight node
	{
		const float *pDeflection = (const float *)&(m_deflectionRows[level][offset][0]);
		const GridWeightNode *pWeightNode = (const GridWeightNode *)pNode;
		float weight = pWeightNode->getWeight();
		float factor2 = 1.0f;
		
		if (m_useWeights)
		{
			int factor = 1 << level;
			
			factor2 = ((float)factor)*((float)factor);
		}
		
		CalculateAddProductC((float *)&(m_subThetas[0]), pDeflection, weight/factor2, m_totalPoints*2, (float *)&(m_tmpBuf[0]));
		if (intensities)
		{
			const float *pAxx = (const float *)&(m_magnificationRows[0][level][offset][0]);
			const float *pAyy = (const float *)&(m_magnificationRows[1][level][offset][0]);
			const float *pAxy = (const float *)&(m_magnificationRows[2][level][offset][0]);

			CalculateAddProductC((float *)&(m_subMag0[0]), pAxx, weight/factor2, m_totalPoints, (float *)&(m_tmpBuf[0]));
			CalculateAddProductC((float *)&(m_subMag1[0]), pAyy, weight/factor2, m_totalPoints, (float *)&(m_tmpBuf[0]));
			CalculateAddProductC((float *)&(m_subMag2[0]), pAxy, weight/factor2, m_totalPoints, (float *)&(m_tmpBuf[0]));
		}
	}
}

void BackProjectTreeBase::calculateBetas(float massFactor)
{
	CopyVector((float *)&(m_betas[0]), (const float *)&(m_thetas[0]), m_totalPoints*2);
	CalculateAddProductC((float *)&(m_betas[0]), (const float *)&(m_subThetas[0]), -massFactor, m_totalPoints*2, (float *)&(m_tmpBuf[0]));
}

void BackProjectTreeBase::calculateMagnifications(float massFactor)
{
	CopyVector((float *)&(m_invMag[0]), (const float *)&(m_oneVector[0]), m_totalPoints);
	CalculateAddProductC((float *)&(m_invMag[0]), (const float *)&(m_subMag0[0]), -massFactor, m_totalPoints, (float *)&(m_tmpBuf2[0]));
	CalculateAddProductC((float *)&(m_invMag[0]), (const float *)&(m_subMag1[0]), massFactor*massFactor, m_totalPoints, (float *)&(m_tmpBuf2[0]));
}

// TODO: move these to derived class?

float BackProjectTreeBase::getTimeDelay(int sourceNumber, int imageNumber, int pointNumber, Vector2D<float> beta) const
{
	return 0.0f;
}

} // end namespace
