/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#ifndef GRALE_GRIDLENSINVERSIONGAFACTORYNEW_H

#define GRALE_GRIDLENSINVERSIONGAFACTORYNEW_H

#include "graleconfig.h"
#include "randomnumbergenerator.h"
#include "backprojectmatrixnew.h"
#include "vector2d.h"
#include "grid.h"
#include <mogal/gafactory.h>
#include <vector>

namespace grale
{

class GridLensInversionGAFactoryParamsNew;
class GridLensInversionGenomeNew;
class ImagesDataExtended;
class GravitationalLens;
class LensFitnessObject;

// NOTE: the virtual inheritance is again very important!
class GRALE_IMPORTEXPORT GridLensInversionGAFactoryNew : public virtual mogal::GAFactory
{
public:
	GridLensInversionGAFactoryNew();
	~GridLensInversionGAFactoryNew();

	mogal::GAFactoryParams *createParamsInstance() const;
	const mogal::GAFactoryParams *getCurrentParameters() const;

	bool init(const mogal::GAFactoryParams *pParams);

	mogal::Genome *createNewGenome() const;

	size_t getMaximalFitnessSize() const							{ return sizeof(float)*(1+getNumberOfFitnessComponents()); } // one for scale factor, some for fitness
	size_t getMaximalGenomeSize() const							{ return (m_numMasses+1)*sizeof(float); } // one extra for the mass sheet value

	bool writeGenome(serut::SerializationInterface &si, const mogal::Genome *pGenome) const;
	bool readGenome(serut::SerializationInterface &si, mogal::Genome **pGenome) const;
	bool writeGenomeFitness(serut::SerializationInterface &si, const mogal::Genome *pGenome) const;
	bool readGenomeFitness(serut::SerializationInterface &si, mogal::Genome *pGenome) const;
	bool writeCommonGenerationInfo(serut::SerializationInterface &si) const;
	bool readCommonGenerationInfo(serut::SerializationInterface &si);

	bool hasFloatingPointFitnessValues() const 						{ return true; }

	const RandomNumberGenerator *getRandomNumberGenerator() const				{ return &m_rndGen; }

	GravitationalLens *createLens(const std::vector<float> &masses, float sheetValue, float scaleFactor, double *pTotalMass, std::string &errStr) const;

	DeflectionMatrix *getDeflectionMatrix()							{ return m_pDeflectionMatrix; }
	BackProjectMatrixNew *getShortBackProjectMatrix()					{ return m_pShortBPMatrix; }
	BackProjectMatrixNew *getTotalBackProjectMatrix()					{ return m_pTotalBPMatrix; }
	LensFitnessObject *getFitnessObject()							{ return m_pFitnessObject; }

	Vector2D<float> getBottomLeft() const							{ return m_bottomLeftFloat; }
	Vector2D<float> getTopRight() const							{ return m_topRightFloat; }
	const std::vector<Vector2D<float> > &getBasisCenters() const				{ return m_basisCenters; }
	float getMinimalMutationWidth() const							{ return m_minMutationWidth; }
	float getMaximalMutationWidth() const							{ return m_maxMutationWidth; }

	float getSheetScale() const								{ return m_sheetScale; }

	virtual bool useAbsoluteMutation() = 0;
	virtual float getMutationAmplitude() = 0;
protected:
	int getMaximumNumberOfGenerations() const						{ return m_maxGenerations; }

	virtual LensFitnessObject *createFitnessObject() = 0;
	virtual bool subInit(LensFitnessObject *pFitnessObject) = 0;
	void onSortedPopulation(const std::vector<mogal::GenomeWrapper> &population);
private:
	bool localSubInit(double z_d, const std::list<ImagesDataExtended *> &images, 
	                  const std::vector<std::pair<GravitationalLens *, Vector2D<double> > > &basisLenses,
			  const GravitationalLens *pBaseLens, bool useSheet);
	double getAngularScale() const								{ return m_pShortBPMatrix->getAngularScale(); }
	bool createBasisLenses(double D_d, const std::list<GridSquare> &squares, 
			       const std::vector<double> &massWeights, double massScale,
			       std::vector<std::pair<GravitationalLens *, Vector2D<double> > > &basisLenses);
	bool recalculateMassWeights(const std::vector<std::pair<GravitationalLens *, Vector2D<double> > > &basisLenses, 
			            std::vector<double> &massWeights, const std::list<GridSquare> &squares, 
				    double massScale);
	void calculateMassRegion(const std::list<GridSquare> &squares, double &minSize);

	RandomNumberGenerator m_rndGen;
	GridLensInversionGAFactoryParamsNew *m_pCurrentParams;
	int m_numMasses, m_maxGenerations;
	bool m_showEvolution, m_saveFirstGenome;
	bool m_allowNegativeValues;
	bool m_useGenomeSheet;
	float m_sheetScale;

	Vector2D<float> m_bottomLeftFloat;
	Vector2D<float> m_topRightFloat;
	Vector2D<double> m_bottomLeft;
	Vector2D<double> m_topRight;
	std::vector<Vector2D<float> > m_basisCenters;
	float m_minMutationWidth;
	float m_maxMutationWidth;

	std::vector<std::pair<GravitationalLens *, Vector2D<double> > > m_basisLenses;

	LensFitnessObject *m_pFitnessObject;
	DeflectionMatrix *m_pDeflectionMatrix;
	BackProjectMatrixNew *m_pShortBPMatrix, *m_pTotalBPMatrix;
};
	
} // end namespace

#endif // GRALE_GRIDLENSINVERSIONGAFACTORYNEW_H

