/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "imageplane.h"
#include "lensplane.h"
#include "sourceplane.h"
#include "sourceimage.h"
#include "gravitationallens.h"
#include "renderserver.h"
#include "triangle2d.h"
#include <enut/tcpv4socket.h>
#include <enut/tcppacketsocket.h>
#include <enut/packet.h>
#include <serut/fileserializer.h>
#include <serut/dummyserializer.h>
#include <serut/memoryserializer.h>
//#include <sys/socket.h>
//#include <netinet/in.h>
//#include <arpa/inet.h>
//#include <sys/time.h>
#include <string.h>
#include <iostream>
//#include <unistd.h>
#include <time.h>

#include "debugnew.h"


namespace grale
{

ImagePlane::ImagePlane(const GravitationalLens *pLens, SourcePlane *s) /* : m_uniDist(&m_rndGen) */
{
	if (pLens)
		m_pLens = pLens->createCopy(); // should work if the lens is valid
	else
		m_pLens = 0;

	sourceplane = s;
	numx = 0;
	numy = 0;
	xstep = 0;
	ystep = 0;
	m_pAlphaxxFunction = 0;
	m_pAlphayyFunction = 0;
	m_pAlphaxyFunction = 0;
}

ImagePlane::~ImagePlane()
{
	if (m_pLens)
		delete m_pLens;
	if (m_pAlphaxxFunction != 0)
		delete m_pAlphaxxFunction;
	if (m_pAlphayyFunction != 0)
		delete m_pAlphayyFunction;
	if (m_pAlphaxyFunction != 0)
		delete m_pAlphaxyFunction;

	std::list<std::list<Vector2D<double> > *>::const_iterator it;

	for (it = caustics2.begin() ; it != caustics2.end() ; it++)
		delete *it;
	for (it = criticallines2.begin() ; it != criticallines2.end() ; it++)
		delete *it;
}

//
//TODO: Avoid copying the lens twice in these functions!
//

bool ImagePlane::init(Vector2D<double> bl,Vector2D<double> tr,int xp,int yp, OpenCLKernel *pCLKernel)
{
	InternalLensPlane lensPlane(m_pLens, (ImagePlane *)this);

	setFeedbackStatus("Calculating lens plane");

	if (!lensPlane.init(bl, tr, xp, yp, pCLKernel))
	{
		setErrorString(lensPlane.getErrorString());
		return false;
	}

	if (!init(&lensPlane))
		return false;

	return true;
}

bool ImagePlane::init(Vector2D<double> bl,Vector2D<double> tr,int xp,int yp, 
		      const nut::NetworkLayerAddress &serverAddress,uint16_t serverport)
{
	InternalLensPlane lensPlane(m_pLens, (ImagePlane *)this);

	setFeedbackStatus("Calculating lens plane");

	if (!lensPlane.init(bl, tr, xp, yp, serverAddress, serverport))
	{
		setErrorString(lensPlane.getErrorString());
		return false;
	}

	if (!init(&lensPlane))
		return false;

	return true;
}

bool ImagePlane::init(LensPlane *lensplane)
{
	if (!lensplane->isInit())
	{
		setErrorString("Can't use an uninitialized lens plane");
		return false;
	}

	if (m_pLens)
		delete m_pLens;
	if (m_pAlphaxxFunction != 0)
		delete m_pAlphaxxFunction;
	if (m_pAlphayyFunction != 0)
		delete m_pAlphayyFunction;
	if (m_pAlphaxyFunction != 0)
		delete m_pAlphaxyFunction;
	
	m_pLens = lensplane->getLens()->createCopy();
	bottomleft = lensplane->getBottomLeft();
	topright = lensplane->getTopRight();
	xstep = lensplane->getXStep();
	ystep = lensplane->getYStep();
	numx = lensplane->getNumXPoints();
	numy = lensplane->getNumYPoints();

	m_betas.resize(numx*numy);
	m_alphaxx.resize(numx*numy);
	m_alphayy.resize(numx*numy);
	m_alphaxy.resize(numx*numy);
	double *inversemagnifications = new double[numx*numy];

	double D_s = sourceplane->getD_s();
	double D_ds = sourceplane->getD_ds();
	double factor = D_ds/D_s;

	setFeedbackStatus("Creating map from lens plane");
	
	for (int y = 0 ; y < numy ; y++)
	{
		for (int x = 0 ; x < numx ; x++)
		{
			Vector2D<double> theta = getIndexCoordinate(x,y);
			m_betas[x+y*numx] = theta - factor*lensplane->getAlpha(x,y);
			
			double axx,ayy,axy;
			
			lensplane->getAlphaDerivatives(x,y,axx,ayy,axy);
			axx *= factor;
			ayy *= factor;
			axy *= factor;
			
			m_alphaxx[x+y*numx] = axx;
			m_alphayy[x+y*numx] = ayy;
			m_alphaxy[x+y*numx] = axy;
			inversemagnifications[x+y*numx] = (1.0-axx)*(1.0-ayy)-axy*axy;
		}
	}

	searchCritCaust(inversemagnifications);

	delete [] inversemagnifications;
	
	m_pAlphaxxFunction = new GridFunction(&(m_alphaxx[0]), bottomleft, topright, numx, numy, false);
	m_pAlphayyFunction = new GridFunction(&(m_alphayy[0]), bottomleft, topright, numx, numy, false);
	m_pAlphaxyFunction = new GridFunction(&(m_alphaxy[0]), bottomleft, topright, numx, numy, false);
	
	setFeedbackStatus("Finished");
	return true;	
}

void ImagePlane::searchCritCaust(double *inversemagnifications)
{	
	int numtotal = numx*numy;
	int index = 0;
	int x,y;
//	std::list<Vector2D<double> > criticalpoints;
	
	// create criticallines list and caustics list

	setFeedbackStatus("Searching for critical lines and caustics");
	
	bool *map1 = new bool[numtotal];
	Vector2D<double> *map2 = new Vector2D<double>[numtotal];
	bool *map3 = new bool[numtotal];

	for (index = 0 ; index < numtotal ; index++)
	{
		map1[index] = false;
		map3[index] = false;
	}
	
	int count = 0;
	int totalcount = (numx-1)*(numy-1);
	for (y = 0 ; y < numy-1 ; y++)
	{
		for (x = 0 ; x < numx-1 ; x++)
		{	
			int index = x+y*numx;
			int index2 = x+(y+1)*numx;
			double inv1,inv2;
			
			if (inversemagnifications[index]*inversemagnifications[index+1] < 0)
			{
				inv1 = inversemagnifications[index];
				inv2 = inversemagnifications[index+1];
			
				double rico = (inv2-inv1)/xstep;
				Vector2D<double> pos = getIndexCoordinate(index%numx,index/numx);
				Vector2D<double> zeropos = Vector2D<double>(pos.getX()-inv1/rico,pos.getY());
				
				int x2,y2,index3;
				
				getIndices(zeropos,&x2,&y2);
				index3 = x2+y2*numx;
				if (!map1[index3])
				{
					map1[index3] = true;
					map2[index3] = zeropos;
				}
			}
			if (inversemagnifications[index]*inversemagnifications[index2] < 0)
			{
				inv1 = inversemagnifications[index];
				inv2 = inversemagnifications[index2];
			
				double rico = (inv2-inv1)/ystep;
				Vector2D<double> pos = getIndexCoordinate(index%numx,index/numx);
				Vector2D<double> zeropos = Vector2D<double>(pos.getX(),pos.getY()-inv1/rico);
				
				int x2,y2,index3;
				
				getIndices(zeropos,&x2,&y2);
				index3 = x2+y2*numx;

				if (!map1[index3])
				{
					map1[index3] = true;
					map2[index3] = zeropos;
				}
			}
			count++;
		}
		
		int pct = (int)(((((double)count)/((double)(totalcount-1))))*100.0+0.5);
		setFeedbackPercentage(pct);
	}

	std::list<std::list<Vector2D<double> > *>::const_iterator it;
	std::list<IntLine2D> lines;

	criticallines.clear();
	caustics.clear();
	for (it = caustics2.begin() ; it != caustics2.end() ; it++)
		delete *it;
	caustics2.clear();

	for (it = criticallines2.begin() ; it != criticallines2.end() ; it++)
		delete *it;
	criticallines2.clear();
	
	index = 0;
	for (y = 0 ; y < numy ; y++)
	{
		for (x = 0 ; x < numx ; x++)
		{
			if (map1[index])
			{
				if ((x < numx-1) && map1[index+1] && !map3[index+1])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x+1,y)));
				if ((x > 0) && map1[index-1] && !map3[index-1])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x-1,y)));
				if ((y > 0) && map1[index-numx] && !map3[index-numx])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x,y-1)));
				if ((y < numy-1) && map1[index+numx] && !map3[index+numx])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x,y+1)));
				if ((x > 0 && y > 0) && map1[index-1-numx] && !map3[index-1-numx])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x-1,y-1)));
				if ((x < numx-1 && y > 0) && map1[index+1-numx] && !map3[index+1-numx])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x+1,y-1)));
				if ((x < numx-1 && y < numy-1) && map1[index+1+numx] && !map3[index+1+numx])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x+1,y+1)));
				if ((x > 0 && y < numy-1) && map1[index-1+numx] && !map3[index-1+numx])
					lines.push_back(IntLine2D(IntVector2D(x,y),IntVector2D(x-1,y+1)));
				map3[index] = true;
			}
			index++;
		}
	}

	while (!lines.empty())
	{
		std::list<IntVector2D> vertices;

		IntLine2D line = *(lines.begin());
		lines.pop_front();

		vertices.push_back(line.getVertex1());
		vertices.push_back(line.getVertex2());

		bool done = false;
		while (!done)
		{
			IntVector2D vfirst = *(vertices.begin());
			IntVector2D vlast = *(--(vertices.end()));

			std::list<IntLine2D>::iterator it = lines.begin();
			bool found = false;
			while (!found && it != lines.end())
			{
				found = true;
				if (vfirst == (*it).getVertex1())
				{
					vertices.push_front((*it).getVertex2());
					lines.erase(it);
				}
				else if (vfirst == (*it).getVertex2())
				{
					vertices.push_front((*it).getVertex1());
					lines.erase(it);
				}
				else if (vlast == (*it).getVertex1())
				{
					vertices.push_back((*it).getVertex2());
					lines.erase(it);
				}
				else if (vlast == (*it).getVertex2())
				{
					vertices.push_back((*it).getVertex1());
					lines.erase(it);
				}
				else
				{
					found = false;
					it++;
				}
			}

			if (!found)
				done = true;
		}

		std::list<Vector2D<double> > *crit = new std::list<Vector2D<double> >;
		std::list<IntVector2D>::iterator it;
		
		for (it = vertices.begin() ; it != vertices.end() ; it++)
		{
			int index = (*it).getX()+(*it).getY()*numx;
			crit->push_back(map2[index]);
		}

		criticallines2.push_back(crit);
	}

	std::list<std::list<Vector2D<double> > *>::const_iterator it1;
	std::list<Vector2D<double> > *crit,*caust;
	std::list<Vector2D<double> >::const_iterator it2;

	for (it1 = criticallines2.begin() ; it1 != criticallines2.end() ; it1++)
	{
		crit = *it1;
		caust = 0;

		it2 = crit->begin();
		while (it2 != crit->end())
		{
			Vector2D<double> beta,beta1,beta2,beta3,beta4;
			Vector2D<double> theta;
			bool inplane = false;

			theta = *it2;
			criticallines.push_back(theta);

			if (calculateBeta(theta, &beta))
				inplane = true;

			if (inplane)
			{
				if (caust == 0)
				{
					caust = new std::list<Vector2D<double> >;
					caustics2.push_back(caust);
				}
				caust->push_back(beta);
				caustics.push_back(beta);
			}
			else
			{
				caust = 0;
			}
			it2++;
		}
	}
		
	delete [] map1;
	delete [] map2;
	delete [] map3;
}

#define IMAGEPLANEID			0x414C5049

bool ImagePlane::write(serut::SerializationInterface &si,bool writesources) const
{
	if (!isInit())
	{
		setErrorString("Image plane is not initialized");
		return false;
	}

	if (!si.writeInt32(IMAGEPLANEID))
	{
		setErrorString(std::string("Error writing image plane ID: ") + si.getErrorString());
		return false;
	}
	if (!si.writeDoubles(bottomleft.getComponents(), 2))
	{
		setErrorString(std::string("Error writing image plane bottom-left coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.writeDoubles(topright.getComponents(), 2))
	{
		setErrorString(std::string("Error writing image plane top-right coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.writeInt32(numx))
	{
		setErrorString(std::string("Error writing image plane number of X points: ") + si.getErrorString());
		return false;
	}
	if (!si.writeInt32(numy))
	{
		setErrorString(std::string("Error writing image plane number of Y points: ") + si.getErrorString());
		return false;
	}

	int numtotal = numx*numy;
	int i;
	for (i = 0 ; i < numtotal ; i++)
	{
		double array[5];

		array[0] = m_betas[i].getX();
		array[1] = m_betas[i].getY();
		array[2] = m_alphaxx[i];
		array[3] = m_alphayy[i];
		array[4] = m_alphaxy[i];

		if (!si.writeDoubles(array,5))
		{
			setErrorString(std::string("Error writing image plane map: ") + si.getErrorString());
			return false;
		}
	}

	if (!m_pLens->write(si))
	{
		setErrorString(std::string("Error writing the image plane's associated lens: ") + m_pLens->getErrorString());
		return false;
	}
	if (!sourceplane->write(si,writesources))
	{
		setErrorString(std::string("Error writing the image plane's associated source plane: ") + sourceplane->getErrorString());
		return false;
	}
	return true;
}

bool ImagePlane::read(serut::SerializationInterface &si,ImagePlane **ip,std::string &errstr)
{
	int32_t id;
	Vector2D<double> bottomleft,topright;
	int32_t numx,numy;
	std::vector<Vector2D<double> > betas;
	std::vector<double> alphaxx;
	std::vector<double> alphayy;
	std::vector<double> alphaxy;
	double *inversemagnifications;
	
	if (!si.readInt32(&id))
	{
		errstr = std::string(std::string("Error reading image plane ID: ") + si.getErrorString());
		return false;
	}
	if (id != IMAGEPLANEID)
	{
		errstr = std::string("Read invalid image plane ID");
		return false;
	}
	if (!si.readDoubles(bottomleft.getComponents(), 2))
	{
		errstr = std::string(std::string("Error reading image plane bottom-left coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.readDoubles(topright.getComponents(), 2))
	{
		errstr = std::string(std::string("Error reading image plane top-right coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.readInt32(&numx))
	{
		errstr = std::string(std::string("Error reading image plane number of X points: ") + si.getErrorString());
		return false;
	}
	if (!si.readInt32(&numy))
	{
		errstr = std::string(std::string("Error reading image plane number of Y points: ") + si.getErrorString());
		return false;
	}

	int32_t numtotal = numx*numy;
	int32_t i;

	betas.resize(numtotal);
	alphaxx.resize(numtotal);
	alphayy.resize(numtotal);
	alphaxy.resize(numtotal);
	inversemagnifications = new double[numtotal];
	
	for (i = 0 ; i < numtotal ; i++)
	{
		double array[5];

		if (!si.readDoubles(array,5))
		{
			delete [] inversemagnifications;
			errstr = std::string(std::string("Error reading image plane map: ") + si.getErrorString());
			return false;
		}
		betas[i] = Vector2D<double>(array[0],array[1]);
		alphaxx[i] = array[2];
		alphayy[i] = array[3];
		alphaxy[i] = array[4];
		inversemagnifications[i] = (1.0-array[2])*(1.0-array[3])-array[4]*array[4];
	}
	
	GravitationalLens *lens;
	SourcePlane *sourceplane;
	std::string errstr2;

	if (!GravitationalLens::read(si,&lens,errstr2))
	{
		delete [] inversemagnifications;
		errstr = std::string(std::string("Error reading the image plane's associated lens: ") + errstr2);
		return false;
	}

	if (!SourcePlane::read(si,&sourceplane,errstr2))
	{
		delete [] inversemagnifications;
		delete lens;
		errstr = std::string(std::string("Error reading the image plane's associated source plane: ") + errstr2);
		return false;
	}
	
	double width,height;

	ImagePlane *plane = new ImagePlane();
	
	width = topright.getX()-bottomleft.getX();
	height = topright.getY()-bottomleft.getY();
	
	plane->xstep = width/((double)(numx-1));
	plane->ystep = height/((double)(numy-1));
	plane->numx = numx;
	plane->numy = numy;
	plane->bottomleft = bottomleft;
	plane->topright = topright;
	plane->m_betas = betas;
	plane->m_alphaxx = alphaxx;
	plane->m_alphayy = alphayy;
	plane->m_alphaxy = alphaxy;
	plane->sourceplane = sourceplane;
	plane->m_pLens = lens;
	plane->m_pAlphaxxFunction = new GridFunction(&(plane->m_alphaxx[0]), bottomleft, topright, numx, numy, false);
	plane->m_pAlphayyFunction = new GridFunction(&(plane->m_alphayy[0]), bottomleft, topright, numx, numy, false);
	plane->m_pAlphaxyFunction = new GridFunction(&(plane->m_alphaxy[0]), bottomleft, topright, numx, numy, false);
	
	plane->searchCritCaust(inversemagnifications);

	delete [] inversemagnifications;
	
	*ip = plane;
	
	return true;
}

bool ImagePlane::load(const std::string &fname,ImagePlane **ip,std::string &errstr)
{
	serut::FileSerializer fs;

	if (!fs.open(fname, serut::FileSerializer::ReadOnly))
	{
		errstr = std::string("Error opening image plane file for reading: ") + fs.getErrorString();
		return false;
	}
	return ImagePlane::read(fs,ip,errstr);
}

bool ImagePlane::save(const std::string &fname,bool writesources) const
{
	serut::FileSerializer fs;
	bool status;

	if (!fs.open(fname, serut::FileSerializer::WriteOnly))
	{
		setErrorString(std::string("Error opening image plane file for writing: ") + fs.getErrorString());
		return false;
	}
	return write(fs,writesources);
}

/*
double ImagePlane::getImageIntensityAccurate(int xPixel, int yPixel, int subSamples) const
{
	int index[4];
	double axx[4];
	double ayy[4];
	double axy[4];
	Vector2D<double> beta[4];

	index[0] = xPixel + yPixel*numx;
	index[1] = index[0] + 1;
	index[2] = index[0] + numx;
	index[3] = index[2] + 1;

	for (int i = 0 ; i < 4 ; i++)
	{
		beta[i] = betas[index[i]];
		axx[i] = alphaxx[index[i]];
		axy[i] = alphaxy[index[i]];
		ayy[i] = alphayy[index[i]];
	}

	Vector2D<double> v1 = (beta[0]-beta[3])/2.0;
	Vector2D<double> v2 = (beta[1]-beta[2])/2.0;
	Vector2D<double> b = (beta[0]+beta[1]+beta[2]+beta[3])/4.0;

	if (!sourceplane->isSourceInRange(b, MAX(v1.getLength(), v2.getLength())*1.1))
		return 0;

	int samp1 = subSamples/2;
	int samp2 = subSamples-subSamples/2;

	double part1 = 0;

	if (samp1 > 0)
	{
		Triangle2D<double> t(beta[0], beta[1], beta[2]);

		part1 = getTriangleIntensity(beta[0], beta[1], beta[2], axx[0], axx[1], axx[2],
			            ayy[0], ayy[1], ayy[2], axy[0], axy[1], axy[2], samp1, true);

		part1 /= (double)samp1;
		part1 *= t.getArea();
	}

	double part2 = 0;

	if (samp2 > 0)
	{
		Triangle2D<double> t(beta[3], beta[1], beta[2]);

		part2 = getTriangleIntensity(beta[3], beta[1], beta[2], axx[3], axx[1], axx[2],
			            ayy[3], ayy[1], ayy[2], axy[3], axy[1], axy[2], samp2, true);

		part2 /= (double)samp2;
		part2 *= t.getArea();
	}

	return (part1+part2)/(xstep*ystep);
}

double ImagePlane::getTriangleIntensity(Vector2D<double> p1, Vector2D<double> p2, Vector2D<double> p3,
			            double axx1, double axx2, double axx3,
				    double ayy1, double ayy2, double ayy3,
				    double axy1, double axy2, double axy3,
				    int numPoints, bool useMagnification) const
{
	Vector2D<double> v1 = p2-p1;
	Vector2D<double> v2 = p3-p1;
	Vector2D<double> vSum = v1+v2;
	double daxx1 = axx2-axx1;
	double daxx2 = axx3-axx1;
	double dayy1 = ayy2-ayy1;
	double dayy2 = ayy3-ayy1;
	double daxy1 = axy2-axy1;
	double daxy2 = axy3-axy1;
	double sum = 0;

	for (int i = 0 ; i < numPoints ; i++)
	{
		double u = m_uniDist.pickNumber();
		double w = m_uniDist.pickNumber();

		Vector2D<double> p = u*v1 + w*v2;
		Vector2D<double> p2 = vSum - p;

		if (p2.getLengthSquared() < p.getLengthSquared())
		{
			u = 1.0-u;
			w = 1.0-w;
			p = u*v1 + w*v2;
		}

		double inverseMagnification = 1;

		if (useMagnification)
		{
			double axx = axx1 + daxx1*u + daxx2*w;
			double ayy = ayy1 + dayy1*u + dayy2*w;
			double axy = axy1 + daxy1*u + daxy2*w;

			inverseMagnification = ABS((1.0-axx)*(1.0-ayy)-axy*axy);
		}

		p += p1;

		sum += sourceplane->getIntensity(p)/inverseMagnification;
	}

	return sum;
}

double ImagePlane::getSourceIntensityAccurate(int xPixel, int yPixel, int subSamples) const
{
	Vector2D<double> beta[4];

	beta[0] = getIndexCoordinate(xPixel, yPixel);
	beta[1] = getIndexCoordinate(xPixel+1, yPixel);
	beta[2] = getIndexCoordinate(xPixel, yPixel+1);
	beta[3] = getIndexCoordinate(xPixel+1, yPixel+1);

	Vector2D<double> v1 = (beta[0]-beta[3])/2.0;
	Vector2D<double> v2 = (beta[1]-beta[2])/2.0;
	Vector2D<double> b = (beta[0]+beta[1]+beta[2]+beta[3])/4.0;

	if (!sourceplane->isSourceInRange(b, MAX(v1.getLength(), v2.getLength())*1.1))
		return 0;

	int samp1 = subSamples/2;
	int samp2 = subSamples-subSamples/2;

	double part1 = 0;

	if (samp1 > 0)
	{
		Triangle2D<double> t(beta[0], beta[1], beta[2]);

		part1 = getTriangleIntensity(beta[0], beta[1], beta[2], 0, 0, 0,
			            0, 0, 0, 0, 0, 0, samp1, false);

		part1 /= (double)samp1;
		part1 *= t.getArea();
	}

	double part2 = 0;

	if (samp2 > 0)
	{
		Triangle2D<double> t(beta[3], beta[1], beta[2]);

		part2 = getTriangleIntensity(beta[3], beta[1], beta[2], 0, 0, 0,
			            0, 0, 0, 0, 0, 0, samp2, false);

		part2 /= (double)samp2;
		part2 *= t.getArea();
	}

	return (part1+part2)/(xstep*ystep);
}
*/

double ImagePlane::getImageIntensityAccurate(int xPixel, int yPixel, int subSamples) const
{
	int index[4];
	double axx[4];
	double ayy[4];
	double axy[4];
	Vector2D<double> beta[4];
	Vector2D<double> theta[4];
	
	theta[0] = getIndexCoordinate(xPixel, yPixel);
	theta[1] = getIndexCoordinate(xPixel + 1, yPixel);
	theta[2] = getIndexCoordinate(xPixel, yPixel + 1);
	theta[3] = getIndexCoordinate(xPixel + 1, yPixel + 1);

	index[0] = xPixel + yPixel*numx;
	index[1] = index[0] + 1;
	index[2] = index[0] + numx;
	index[3] = index[2] + 1;

	for (int i = 0 ; i < 4 ; i++)
	{
		beta[i] = m_betas[index[i]];
		axx[i] = m_alphaxx[index[i]];
		axy[i] = m_alphaxy[index[i]];
		ayy[i] = m_alphayy[index[i]];
	}

	Vector2D<double> v1 = (beta[0]-beta[3])/2.0;
	Vector2D<double> v2 = (beta[1]-beta[2])/2.0;
	Vector2D<double> b = (beta[0]+beta[1]+beta[2]+beta[3])/4.0;

	if (!sourceplane->isSourceInRange(b, MAX(v1.getLength(), v2.getLength())*1.1))
		return 0;

	// First we'll calculate the surface brightness, using subsampling if requested
	
	int numSub = (int)(SQRT((double)subSamples)+0.5);

	if (numSub < 1)
		numSub = 1;

	double sum = 0;
	double ySubStep = ystep/(double)numSub;
	double xSubStep = xstep/(double)numSub;

	for (int i = 0 ; i < numSub ; i++)
	{
		double dy = ((double)i)*ySubStep + ySubStep/2.0;
		double thetay = theta[0].getY() + dy;

		for (int j = 0 ; j < numSub ; j++)
		{
			double dx = ((double)j)*xSubStep + xSubStep/2.0;
			double thetax = theta[0].getX() + dx;

			Vector2D<double> b[4];

			for (int k = 0 ; k < 4 ; k++)
			{
				double dthetax = thetax - theta[k].getX();
				double dthetay = thetay - theta[k].getY();

				double bx = beta[k].getX() + (1.0 - axx[k])*dthetax - axy[k]*dthetay;
				double by = beta[k].getY() - axy[k]*dthetax + (1.0 - ayy[k])*dthetay;
				
				b[k] = Vector2D<double>(bx, by);
			}
			
			double u = dx/xstep;
			double v = dy/ystep;

			Vector2D<double> B = b[0]*(1.0-u)*(1.0-v) + b[1]*u*(1.0-v) + b[2]*(1.0-u)*v + b[3]*u*v;

			sum += sourceplane->getSurfaceBrightness(B); // surface brightness is conserved
		}
	}

	sum /= (double)(numSub*numSub);
	
	// Then we'll process the point sources
	
	std::list<Vector2DPlus> pointSourceInfo;
	std::list<Vector2DPlus>::const_iterator it;
	
	int refIdx[2] = {0, 3};

	for (int i = 0 ; i < 2 ; i++)
	{
		int k = refIdx[i];
		
		sourceplane->getPointSourceIntensities(Triangle2D<double>(beta[k], beta[1], beta[2]), pointSourceInfo);
		
		Vector2D<double> p = beta[1]-beta[k];
		Vector2D<double> q = beta[2]-beta[k];
		double daxx1 = axx[1]-axx[k];
		double daxx2 = axx[2]-axx[k];
		double dayy1 = ayy[1]-ayy[k];
		double dayy2 = ayy[2]-ayy[k];
		double daxy1 = axy[1]-axy[k];
		double daxy2 = axy[2]-axy[k];

		double det = p.getX()*q.getY()-p.getY()*q.getX();
		double Ax = q.getY()/det;
		double Ay = -q.getX()/det;
		double Bx = -p.getY()/det;
		double By = p.getX()/det;
		
		for (it = pointSourceInfo.begin() ; it != pointSourceInfo.end() ; it++)
		{
			double dx = (*it).getX() - beta[k].getX();
			double dy = (*it).getY() - beta[k].getY();

			double u = Ax*dx+Ay*dy;
			double w = Bx*dx+By*dy;
			
			double axx0 = axx[k] + daxx1*u + daxx2*w;
			double ayy0 = ayy[k] + dayy1*u + dayy2*w;
			double axy0 = axy[k] + daxy1*u + daxy2*w;

			double inverseMagnification = ABS((1.0-axx0)*(1.0-ayy0)-axy0*axy0);

			sum += (*it).getValue()/inverseMagnification;
		}
	}
	
	return sum;
}

double ImagePlane::getSourceIntensityAccurate(int xPixel, int yPixel, int subSamples) const
{
	int index[4];
	Vector2D<double> theta[4];
	
	theta[0] = getIndexCoordinate(xPixel, yPixel);
	theta[1] = getIndexCoordinate(xPixel + 1, yPixel);
	theta[2] = getIndexCoordinate(xPixel, yPixel + 1);
	theta[3] = getIndexCoordinate(xPixel + 1, yPixel + 1);

	Vector2D<double> v1 = (theta[0]-theta[3])/2.0;
	Vector2D<double> v2 = (theta[1]-theta[2])/2.0;
	Vector2D<double> t = (theta[0]+theta[1]+theta[2]+theta[3])/4.0;

	if (!sourceplane->isSourceInRange(t, MAX(v1.getLength(), v2.getLength())*1.1))
		return 0;

	// First we'll calculate the surface brightness, using subsampling if requested
	
	int numSub = (int)(SQRT((double)subSamples)+0.5);

	if (numSub < 1)
		numSub = 1;

	double sum = 0;
	double ySubStep = ystep/(double)numSub;
	double xSubStep = xstep/(double)numSub;

	for (int i = 0 ; i < numSub ; i++)
	{
		double dy = ((double)i)*ySubStep + ySubStep/2.0;
		double thetay = theta[0].getY() + dy;

		for (int j = 0 ; j < numSub ; j++)
		{
			double dx = ((double)j)*xSubStep + xSubStep/2.0;
			double thetax = theta[0].getX() + dx;

			sum += sourceplane->getSurfaceBrightness(Vector2D<double>(thetax, thetay));
		}
	}

	sum /= (double)(numSub*numSub);
	
	// Then we'll process the point sources
	
	std::list<Vector2DPlus> pointSourceInfo;
	std::list<Vector2DPlus>::const_iterator it;
	
	int refIdx[2] = {0, 3};

	for (int i = 0 ; i < 2 ; i++)
	{
		int k = refIdx[i];
		
		sourceplane->getPointSourceIntensities(Triangle2D<double>(theta[k], theta[1], theta[2]), pointSourceInfo);
		
		for (it = pointSourceInfo.begin() ; it != pointSourceInfo.end() ; it++)
			sum += (*it).getValue();
	}
	
	return sum;
}

bool ImagePlane::traceBeta(Vector2D<double> beta, std::list<Vector2D<double> > &thetaPoints, int numIterations, bool check) const
{
	thetaPoints.clear();

	for (int y = 0 ; y < numy-1 ; y++)
	{
		for (int x = 0 ; x < numx-1 ; x++)
		{
			Vector2D<double> b1 = traceIndexCoordinate(x, y);
			Vector2D<double> b2 = traceIndexCoordinate(x+1, y);
			Vector2D<double> b3 = traceIndexCoordinate(x, y+1);
			Vector2D<double> b4 = traceIndexCoordinate(x+1, y+1);
			Vector2D<double> t1 = getIndexCoordinate(x, y);
			Vector2D<double> t2 = getIndexCoordinate(x+1, y);
			Vector2D<double> t3 = getIndexCoordinate(x, y+1);
			Vector2D<double> t4 = getIndexCoordinate(x+1, y+1);

			Triangle2D<double> triangle1(b1, b2, b3);
			Triangle2D<double> triangle2(b4, b2, b3);

			if (triangle1.isInside(beta) || triangle2.isInside(beta))
			{
				Vector2D<double> theta;
				
				if (!refinePosition(beta, (t1+t4)/2.0, theta, numIterations))
					return false;

				thetaPoints.push_back(theta);
			}
		}
	}

	if (check)
	{
		std::list<Vector2D<double> >::iterator it;
		double pixelScale = SQRT(xstep*xstep+ystep*ystep);

		it = thetaPoints.begin();
		
		while (it != thetaPoints.end())
		{
			Vector2D<double> theta = *it;
			Vector2D<double> newBeta;
			bool ok = true;
		
			if (!m_pLens->traceTheta(sourceplane->getD_s(), sourceplane->getD_ds(), theta, &newBeta))
				ok = false;
			else
			{
				double invMag = m_pLens->getInverseMagnification(sourceplane->getD_s(), sourceplane->getD_ds(), theta);
				//double lengthResize = SQRT(invMag); // magnification relates areas, we'll take the square root to compare distances
				
				Vector2D<double> diff = newBeta-beta;

				//if (diff.getLength() > lengthResize*pixelScale*2.0) // to be on the safe side
				if (diff.getLength() > pixelScale*2.0) // to be on the safe side
					ok = false;
			}

			if (ok)
				it++;
			else
				it = thetaPoints.erase(it);
		}
	}
	
	return true;
}

bool ImagePlane::refinePosition(grale::Vector2D<double> beta, 
                                grale::Vector2D<double> startTheta,
	 		        grale::Vector2D<double> &theta,
				int numIterations) const
{	
	grale::Vector2D<double> newTheta = startTheta;
	grale::Vector2D<double> newBeta;
	
	double prevDiff = 1e30;
	bool done = false;
		
	if (!m_pLens->traceTheta(sourceplane->getD_s(), sourceplane->getD_ds(), newTheta, &newBeta))
	{
		setErrorString(std::string("Unable to trace a point to the source plane: ") + m_pLens->getErrorString());
		return false;
	}

	double fraction = sourceplane->getD_ds()/sourceplane->getD_s();
	grale::Vector2D<double> diffBeta = beta - newBeta;
	grale::Vector2D<double> closestPoint = newTheta;
	prevDiff = diffBeta.getLength();

	for (int i = 0 ; i < numIterations ; i++)
	{
		double mxx, myy, mxy;
		
		if (!m_pLens->getAlphaVectorDerivatives(newTheta, mxx, myy, mxy))
		{
			setErrorString(m_pLens->getErrorString());
			return false;
		}

		mxx = (1.0-fraction*mxx);
		myy = (1.0-fraction*myy);
		mxy = -fraction*mxy;

		double denom = mxx*myy-mxy*mxy;
		double dx = (myy*diffBeta.getX() - mxy*diffBeta.getY())/denom;
		double dy = (-mxy*diffBeta.getX() + mxx*diffBeta.getY())/denom;

		newTheta += grale::Vector2D<double>(dx, dy);
		if (!m_pLens->traceTheta(sourceplane->getD_s(), sourceplane->getD_ds(), newTheta, &newBeta))
		{
			setErrorString(std::string("Unable to trace a point to the source plane: ") + m_pLens->getErrorString());
			return false;
		}

		diffBeta = beta - newBeta;

		if (diffBeta.getLength() < prevDiff)
		{
			prevDiff = diffBeta.getLength();
			closestPoint = newTheta;
		}
	}

	theta = closestPoint;
	return true;
}

bool ImagePlane::traceTheta(Vector2D<double> theta, Vector2D<double> *pBeta) const
{
	if (!m_pLens->traceTheta(sourceplane->getD_s(), sourceplane->getD_ds(), theta, pBeta))
	{
		setErrorString(m_pLens->getErrorString());
		return false;
	}	
	return true;
}

bool ImagePlane::calculateBeta(Vector2D<double> theta, Vector2D<double> *pBeta) const
{
	// TODO: make this configurable
#if 0
	if (!m_pLens->traceTheta(sourceplane->getD_s(), sourceplane->getD_ds(), theta, pBeta))
		return false;
#else
	if (!traceThetaApprox(theta, pBeta))
		return false;
#endif 
	return true;
}

bool ImagePlane::traceThetaApprox(Vector2D<double> theta, Vector2D<double> *pBeta) const
{
	int xIdx, yIdx;

	xIdx = (int)(((theta.getX()-bottomleft.getX())/xstep));
	yIdx = (int)(((theta.getY()-bottomleft.getY())/ystep));

	if (xIdx < 0 || xIdx >= numx-1 || yIdx < 0 || yIdx >= numy-1)
	{
		setErrorString("Theta vector lies outside the image plane");
		return false;
	}

	Vector2D<double> b1 = approximateBeta(xIdx, yIdx, theta);
	Vector2D<double> b2 = approximateBeta(xIdx+1, yIdx, theta);
	Vector2D<double> b3 = approximateBeta(xIdx, yIdx+1, theta);
	Vector2D<double> b4 = approximateBeta(xIdx+1, yIdx+1, theta);
	
	double u = (theta.getX() - bottomleft.getX())/xstep - (double)xIdx;
	double v = (theta.getY() - bottomleft.getY())/ystep - (double)yIdx;

	*pBeta = b1*(1.0-u)*(1.0-v) + b2*u*(1.0-v) + b3*(1.0-u)*v + b4*u*v;

	return true;
}

Vector2D<double> ImagePlane::approximateBeta(int xIdx, int yIdx, Vector2D<double> theta) const
{
	int index = xIdx + yIdx*numx;

	Vector2D<double> beta0 = m_betas[index];
	Vector2D<double> theta0 = getIndexCoordinate(xIdx, yIdx);
	double axx0 = m_alphaxx[index];
	double ayy0 = m_alphayy[index];
	double axy0 = m_alphaxy[index];

	double dthetax = theta.getX() - theta0.getX();
	double dthetay = theta.getY() - theta0.getY();

	double bx = beta0.getX() + (1.0 - axx0)*dthetax - axy0*dthetay;
	double by = beta0.getY() - axy0*dthetax + (1.0 - ayy0)*dthetay;

	return Vector2D<double>(bx, by);
}

double ImagePlane::getInverseMagnification(Vector2D<double> theta) const
{
	return m_pLens->getInverseMagnification(sourceplane->getD_s(), sourceplane->getD_ds(), theta);
}

} // end namespace

