/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "lensplane.h"
#include "gravitationallens.h"
#include "renderserver.h"
#include "compositelens.h"
#include "deflectiongridlens.h"
#include "openclkernel.h"
#include "constants.h"
#include <enut/tcpv4socket.h>
#include <enut/tcppacketsocket.h>
#include <enut/packet.h>
#include <serut/fileserializer.h>
#include <serut/dummyserializer.h>
#include <serut/memoryserializer.h>
#include <string.h>
//#include <unistd.h>
#include <stdlib.h>
#include <time.h>

#include <iostream>

#include "debugnew.h"

namespace grale
{

LensPlane::LensPlane(const GravitationalLens *pLens)
{
	m_pLens = pLens->createCopy(); // should work if pLens is valid
	alphas.resize(0);
	alphaxx.resize(0);
	alphayy.resize(0);
	alphaxy.resize(0);
	numx = 0;
	numy = 0;
	xstep = 0;
	ystep = 0;
	m_init = false;
}

LensPlane::~LensPlane()
{
	delete m_pLens;
}

bool LensPlane::init(Vector2D<double> bl,Vector2D<double> tr,int xp,int yp, OpenCLKernel *pCLKernel)
{
	m_init = false;
	
	double width,height;
	
	if (xp < 2 || yp < 2)
	{
		setErrorString("Number of X and Y points should be at least 2");
		return false;
	}
	if (xp > 100000 && yp > 100000)
	{
		setErrorString("Number of X and Y points should be at most 100000");
		return false;
	}
	
	width = tr.getX()-bl.getX();
	height = tr.getY()-bl.getY();

	if (width < 0.0L || height < 0.0L)
	{
		setErrorString("Specified coordinates imply a negative width or height");
		return false;
	}
	
	numx = xp;
	numy = yp;

	alphas.resize(numx*numy);
	alphaxx.resize(numx*numy);
	alphayy.resize(numx*numy);
	alphaxy.resize(numx*numy);

	xstep = width/((double)(numx-1));
	ystep = height/((double)(numy-1));

	bottomleft = bl;
	topright = tr;

	double minstep = xstep;

	if (ystep < xstep)
		minstep = ystep;

	m_pLens->setDerivativeAngularDistanceScale(minstep/100.0);

	int numtotal = numx*numy;

#ifdef GRALECONFIG_SUPPORT_OPENCL
	bool calcNative = false;

	if (!pCLKernel)
		calcNative = true;
	else
	{
		setFeedbackStatus("Using GPU version");

		bool renderError = false;

		if (!initUsingGPU(pCLKernel, false, -1, renderError))
		{
			if (renderError)
			{
				setFeedbackStatus("Failed to render everything at once, trying sublenses");

				pCLKernel->destroy();
				if (!pCLKernel->init())
				{
					setFeedbackStatus("Couldn't reinitialize OpenCL kernel");
					calcNative = true;
				}
				
				int totalNum = m_pLens->getCLSubLenses();

				if (totalNum < 0)
				{
					// TODO
					calcNative = true;
					setFeedbackStatus("No sublenses available");
				}
				else
				{
					// For testing: we'll use 1 sublens at a time

					if (!initUsingGPU(pCLKernel, false, 1, renderError))
					{
						calcNative = true;
					}
				}
			}
			else
				calcNative = true;
		}
	}
	
	if (calcNative)
	{
		setFeedbackStatus("Using CPU version");
#endif // GRALECONFIG_SUPPORT_OPENCL

		// fill in the arrays
		
		int index = 0;
		
		setFeedbackStatus("Building map");
		
		for (int y = 0 ; y < numy ; y++)
		{
			for (int x = 0 ; x < numx ; x++, index++)
			{
				Vector2D<double> theta = getIndexCoordinate(x,y);
				Vector2D<double> a;
				double axx,ayy,axy;
				
				if (!m_pLens->getAlphaVector(theta,&a))
				{
					setErrorString(std::string("Error raytracing theta vector: ") + m_pLens->getErrorString());
					return false;
				}
				
				alphas[index] = a;
				m_pLens->getAlphaVectorDerivatives(theta,axx,ayy,axy);
				alphaxx[index] = axx;
				alphayy[index] = ayy;
				alphaxy[index] = axy;
			}
			
			int pct = (int)(((((double)index)/((double)(numtotal-1))))*100.0+0.5);
			setFeedbackPercentage(pct);
		}
#ifdef GRALECONFIG_SUPPORT_OPENCL
	}
#endif // GRALECONFIG_SUPPORT_OPENCL

	setFeedbackStatus("Finished");

	m_init = true;
	
	return true;
}

bool LensPlane::init(Vector2D<double> bl,Vector2D<double> tr,int xp,int yp,const nut::NetworkLayerAddress &serverAddress,uint16_t serverport)
{
	m_init = false;
	
	double width,height;
	
	if (xp < 2 || yp < 2)
	{
		setErrorString("Number of X and Y points should be at least 2");
		return false;
	}
	if (xp > 100000 && yp > 100000)
	{
		setErrorString("Number of X and Y points should be at most 100000");
		return false;
	}
	
	width = tr.getX()-bl.getX();
	height = tr.getY()-bl.getY();

	if (width < 0.0L || height < 0.0L)
	{
		setErrorString("Specified coordinates imply a negative width or height");
		return false;
	}
	
	numx = xp;
	numy = yp;

	alphas.resize(numx*numy);
	alphaxx.resize(numx*numy);
	alphayy.resize(numx*numy);
	alphaxy.resize(numx*numy);

	xstep = width/((double)(numx-1));
	ystep = height/((double)(numy-1));

	bottomleft = bl;
	topright = tr;
	
	double minstep = xstep;

	if (ystep < xstep)
		minstep = ystep;

	m_pLens->setDerivativeAngularDistanceScale(minstep/100.0);

	// let server do the work
	
	if (!renderRemote(serverAddress,serverport))
	{
		return false;
	}
	
	setFeedbackStatus("Finished");
	m_init = true;
	
	return true;
}

bool LensPlane::renderRemote(const nut::NetworkLayerAddress &serverAddress,uint16_t serverPort)
{
	nut::TCPv4Socket tcpSocket;

	setFeedbackStatus("Connecting to server");

	if (!tcpSocket.create())
	{
		setErrorString(std::string("Error creating TCP socket: ") + tcpSocket.getErrorString());
		return false;
	}

	if (!tcpSocket.connect(serverAddress, serverPort))
	{
		setErrorString(std::string("Error opening connection with server: ") + tcpSocket.getErrorString());
		return false;
	}
	
	nut::TCPPacketSocket conn(&tcpSocket, false, false, RENDERSERVER_MAXPACKSIZE, sizeof(uint32_t), RENDERSERVER_PACKID);

	// connection established, send identification
	{
		uint8_t buf[sizeof(int32_t)];
		serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));
		mser.writeInt32(RENDERSERVER_COMMAND_CLIENT);
		
		setFeedbackStatus("Sending identification to server");
		if (!conn.write(buf, sizeof(int32_t)))
		{
			setErrorString(std::string("Error sending identification to server: ") + conn.getErrorString());
			return false;
		}
	}

	setFeedbackStatus("Waiting for reply from server");
	bool done = false;
	while (!done)
	{
		if (!conn.waitForData(0, 200000))
		{
			setErrorString(conn.getErrorString());
			return false;
		}

		if (conn.isDataAvailable())
		{
			size_t length = 0;

			conn.getAvailableDataLength(length);
			if (length == 0)
			{
				setErrorString("Server closed connection");
				return false;
			}
			if (!conn.poll())
			{
				setErrorString(std::string("Error polling for data: ") + conn.getErrorString());
				return false;
			}

			nut::Packet pack;
			while (!done && conn.read(pack))
			{
				if (pack.getLength() != 4)
				{
					setErrorString("Received invalid reply from server");
					return false;
				}

				serut::MemorySerializer mser(pack.getData(), pack.getLength(), 0, 0);
				int32_t cmd;

				mser.readInt32(&cmd);
				if (cmd == RENDERSERVER_COMMAND_ACCEPT)
				{
					done = true;
				}
				else if (cmd == RENDERSERVER_COMMAND_BUSY)
				{
					setErrorString("Server is already rendering an image plane for another client");
					return false;
				}
				else if (cmd == RENDERSERVER_COMMAND_UNAVAILABLE)
				{
					setErrorString("Server has no connected renderers");
					return false;
				}
				else
				{
					setErrorString("Received invalid reply from server");
					return false;
				}
			}
		}
	}

	setFeedbackStatus("Got positive reply from server, sending data");
	
	// send lens info
	{
		serut::DummySerializer lengthcounter;

		m_pLens->write(lengthcounter);
		int bufferlength = lengthcounter.getBytesWritten() + sizeof(int32_t) + sizeof(Vector2D<double>)*2 + sizeof(int32_t)*2 + sizeof(double)*2;
		uint8_t *buffer = new uint8_t [bufferlength];

		serut::MemorySerializer mser(0,0,buffer,bufferlength);
		mser.writeInt32(RENDERSERVER_COMMAND_LENS);
		mser.writeInt32(numx);
		mser.writeInt32(numy);
		mser.writeDoubles(bottomleft.getComponents(), 2);
		mser.writeDoubles(topright.getComponents(), 2);
		mser.writeDouble(1.0); // D_s
		mser.writeDouble(1.0); // D_ds
		m_pLens->write(mser);
		
		if (!conn.write(buffer, bufferlength))
		{
			delete [] buffer;
			setErrorString(std::string("Error writing lens plane info: ") + conn.getErrorString());
			return false;
		}
		
		delete [] buffer;
	}

	setFeedbackStatus("Rendering...");
	setFeedbackPercentage(0);

	double *resultdata = 0;
	
	done = false;
	while (!done)
	{
		if (!conn.waitForData(0, 200000))
		{
			setErrorString(conn.getErrorString());
			return false;
		}

		if (conn.isDataAvailable())
		{
			size_t length = 0;

			conn.getAvailableDataLength(length);
			if (length == 0)
			{
				setErrorString("Server closed connection");
				return false;
			}
			if (!conn.poll())
			{
				setErrorString(std::string("Error polling for data: ") + conn.getErrorString());
				return false;
			}

			nut::Packet pack;
			while (!done && conn.read(pack))
			{
				if (pack.getLength() < 4)
				{
					setErrorString("Received invalid packet from server");
					return false;
				}
				
				serut::MemorySerializer mser(pack.getData(), pack.getLength(), 0, 0);
				int32_t cmd;

				mser.readInt32(&cmd);

				if (cmd == RENDERSERVER_COMMAND_ECHO)
				{
					// ignore it
				}
				else if (cmd == RENDERSERVER_COMMAND_PCT)
				{
					if (pack.getLength() != 8)
					{
						setErrorString("Received invalid render status update from server");
						return false;
					}
					int32_t pct;

					mser.readInt32(&pct);
					setFeedbackPercentage(pct);
				}
				else if (cmd == RENDERSERVER_COMMAND_IMAGEPLANE)
				{
					int resultbytes = pack.getLength()-sizeof(int32_t);

					if ((resultbytes%sizeof(double)) != 0)
					{
						setErrorString("Received invalid render result from server");
						return false;
					}

					int resultreals = resultbytes/sizeof(double);

					if (resultreals != numx*numy*5)
					{
						setErrorString("Received invalid render result from server");
						return false;
					}
					
					resultdata = new double[resultreals];
					memcpy(resultdata, pack.getData()+sizeof(int32_t), resultbytes);

					done = true;
		
					uint8_t buf[sizeof(int32_t)];
					serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));
					mser.writeInt32(RENDERSERVER_COMMAND_ACKIMAGEPLANE);

					if (!conn.write(buf,sizeof(int32_t)))
					{
						setErrorString(std::string("Error writing lens plane acknowledgement: ") + conn.getErrorString());
						return false;
					}
				}
				else if (cmd == RENDERSERVER_COMMAND_UNAVAILABLE)
				{
					setErrorString("Server has no connected renderers anymore");
					return false;
				}
				else
				{
					setErrorString("Received invalid packet from server");
					return false;
				}
			}
		}

		if (!done)
		{
			time_t curtime = time(0);

			if ((curtime - conn.getLastWriteTime()) > 10)
			{
				uint8_t buf[sizeof(int32_t)];
				serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));
				mser.writeInt32(RENDERSERVER_COMMAND_ECHO);

				if (!conn.write(buf,sizeof(int32_t)))
				{
					setErrorString(std::string("Error writing echo packet: ") + conn.getErrorString());
					return false;
				}
			}

			if ((curtime - conn.getLastReadTime()) > 1200)
			{
				setErrorString("Connection with server timed out");
				return false;
			}
		}
	}
	
	tcpSocket.destroy();

	setFeedbackPercentage(100);
	setFeedbackStatus("Received solution from server");
	setFeedbackStatus("Closed server connection");
	
	int index = 0;
	int index2 = 0;
	for (int y = 0 ; y < numy ; y++)
	{
		for (int x = 0 ; x < numx ; x++)
		{
			alphas[index] = Vector2D<double>(resultdata[index2],resultdata[index2+1]);
			alphaxx[index] = resultdata[index2+2];
			alphayy[index] = resultdata[index2+3];
			alphaxy[index] = resultdata[index2+4];
			
			index++;
			index2 += 5;
		}
	}

	delete [] resultdata;
	
	return true;
}

#define LENSPLANEID			0x414C504C

bool LensPlane::write(serut::SerializationInterface &si) const
{
	if (!isInit())
	{
		setErrorString("Lens plane is not initialized");
		return false;
	}

	if (!si.writeInt32(LENSPLANEID))
	{
		setErrorString(std::string("Error writing lens plane ID: ") + si.getErrorString());
		return false;
	}
	if (!si.writeDoubles(bottomleft.getComponents(), 2))
	{
		setErrorString(std::string("Error writing lens plane bottom-left coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.writeDoubles(topright.getComponents(), 2))
	{
		setErrorString(std::string("Error writing lens plane top-right coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.writeInt32(numx))
	{
		setErrorString(std::string("Error writing lens plane number of X points: ") + si.getErrorString());
		return false;
	}
	if (!si.writeInt32(numy))
	{
		setErrorString(std::string("Error writing lens plane number of Y points: ") + si.getErrorString());
		return false;
	}

	int numtotal = numx*numy;
	int i;
	for (i = 0 ; i < numtotal ; i++)
	{
		double array[5];

		array[0] = alphas[i].getX();
		array[1] = alphas[i].getY();
		array[2] = alphaxx[i];
		array[3] = alphayy[i];
		array[4] = alphaxy[i];

		if (!si.writeDoubles(array,5))
		{
			setErrorString(std::string("Error writing lens plane map: ") + si.getErrorString());
			return false;
		}
	}

	if (!m_pLens->write(si))
	{
		setErrorString(std::string("Error writing the lens plane's associated lens: ") + m_pLens->getErrorString());
		return false;
	}

	return true;
}

bool LensPlane::read(serut::SerializationInterface &si,LensPlane **ip,std::string &errstr)
{
	int32_t id;
	Vector2D<double> bottomleft,topright;
	int32_t numx,numy;
	std::vector<Vector2D<double> > alphas;
	std::vector<double> alphaxx, alphayy, alphaxy;
	
	if (!si.readInt32(&id))
	{
		errstr = std::string(std::string("Error reading lens plane ID: ") + si.getErrorString());
		return false;
	}
	if (id != LENSPLANEID)
	{
		errstr = std::string("Read invalid lens plane ID");
		return false;
	}
	if (!si.readDoubles(bottomleft.getComponents(), 2))
	{
		errstr = std::string(std::string("Error reading lens plane bottom-left coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.readDoubles(topright.getComponents(), 2))
	{
		errstr = std::string(std::string("Error reading lens plane top-right coordinates: ") + si.getErrorString());
		return false;
	}
	if (!si.readInt32(&numx))
	{
		errstr = std::string(std::string("Error reading lens plane number of X points: ") + si.getErrorString());
		return false;
	}
	if (!si.readInt32(&numy))
	{
		errstr = std::string(std::string("Error reading lens plane number of Y points: ") + si.getErrorString());
		return false;
	}

	int32_t numtotal = numx*numy;
	int32_t i;

	alphas.resize(numtotal);
	alphaxx.resize(numtotal);
	alphayy.resize(numtotal);
	alphaxy.resize(numtotal);
	
	for (i = 0 ; i < numtotal ; i++)
	{
		double array[5];

		if (!si.readDoubles(array,5))
		{
			errstr = std::string(std::string("Error reading lens plane map: ") + si.getErrorString());
			return false;
		}
		alphas[i] = Vector2D<double>(array[0],array[1]);
		alphaxx[i] = array[2];
		alphayy[i] = array[3];
		alphaxy[i] = array[4];
	}
	
	GravitationalLens *lens;
	std::string errstr2;

	if (!GravitationalLens::read(si,&lens,errstr2))
	{
		errstr = std::string(std::string("Error reading the lens plane's associated lens: ") + errstr2);
		return false;
	}

	double width,height;

	LensPlane *plane = new LensPlane();
	
	width = topright.getX()-bottomleft.getX();
	height = topright.getY()-bottomleft.getY();
	
	plane->xstep = width/((double)(numx-1));
	plane->ystep = height/((double)(numy-1));
	plane->numx = numx;
	plane->numy = numy;
	plane->bottomleft = bottomleft;
	plane->topright = topright;
	plane->alphas = alphas;
	plane->alphaxx = alphaxx;
	plane->alphayy = alphayy;
	plane->alphaxy = alphaxy;
	plane->m_pLens = lens;
	plane->m_init = true;

	*ip = plane;
	
	return true;
}

bool LensPlane::load(const std::string &fname,LensPlane **ip,std::string &errstr)
{
	serut::FileSerializer fs;
	bool status;

	if (!fs.open(fname, serut::FileSerializer::ReadOnly))
	{
		errstr = std::string("Error opening lens plane file for reading: ") + fs.getErrorString();
		return false;
	}
	return LensPlane::read(fs,ip,errstr);
}

bool LensPlane::save(const std::string &fname) const
{
	serut::FileSerializer fs;
	bool status;

	if (!fs.open(fname, serut::FileSerializer::WriteOnly))
	{
		setErrorString(std::string("Error opening lens plane file for writing: ") + fs.getErrorString());
		return false;
	}
	return write(fs);
}

bool LensPlane::scaleDeflections(double factor)
{
	CompositeLensParams newParams;

	if (!newParams.addLens(factor, Vector2D<double>(0, 0), 0, *m_pLens))
	{
		setErrorString(std::string("Couldn't create scaled version of the lens: ") + newParams.getErrorString());
		return false;
	}

	CompositeLens *pNewLens = new CompositeLens();

	if (!pNewLens->init(m_pLens->getLensDistance(), &newParams))
	{
		setErrorString(std::string("Couldn't create scaled version of the lens: ") + newParams.getErrorString());
		delete pNewLens;
		return false;
	}

	// Ok, got the new lens. Now store it and scale the deflections

	delete m_pLens;
	m_pLens = pNewLens;

	int index = 0;
	for (int y = 0 ; y < numy ; y++)
	{
		for (int x = 0 ; x < numx ; x++)
		{
			alphas[index] *= factor;
			alphaxx[index] *= factor;
			alphayy[index] *= factor;
			alphaxy[index] *= factor;
			
			index++;
		}
	}

	return true;
}

GravitationalLens *LensPlane::createDeflectionGridLens() const
{
	int totalNum = numx*numy;
	std::vector<double> alphaX(totalNum), alphaY(totalNum);

	for (int i = 0 ; i < totalNum ; i++)
	{
		alphaX[i] = alphas[i].getX();
		alphaY[i] = alphas[i].getY();
	}

	DeflectionGridLensParams lensParams(alphaX, alphaY, numx, numy, bottomleft, topright);
	DeflectionGridLens *pNewLens = new DeflectionGridLens();

	if (!pNewLens->init(m_pLens->getLensDistance(), &lensParams))
	{
		setErrorString(std::string("Couldn't initialize deflection grid lens: ") + pNewLens->getErrorString());
		delete pNewLens;
		return 0;
	}

	return pNewLens;
}

bool LensPlane::initUsingGPU(OpenCLKernel *pCLKernel, bool linePerLine, int numSubLenses, bool &renderError)
{
	renderError = false;
#ifndef GRALECONFIG_SUPPORT_OPENCL
	return false;
#else
	int numIntParams;
	int numFloatParams;

	if (!m_pLens->getCLParameterCounts(&numIntParams, &numFloatParams))
	{
		setFeedbackStatus(std::string("Couldn't get parameter counts for OpenCL program: ") + m_pLens->getErrorString());
		return false;
	}
	double deflectionScale, potentialScale;

	if (!m_pLens->getSuggestedScales(&deflectionScale, &potentialScale))
	{
		deflectionScale = ANGLE_ARCSEC;
		potentialScale = ANGLE_ARCSEC*ANGLE_ARCSEC;
	}

	std::vector<int> intParams(numIntParams+1);
	std::vector<float> floatParams(numFloatParams+1);

	if (!m_pLens->getCLParameters(deflectionScale, potentialScale, &(intParams[0]), &(floatParams[0])))
	{
		setFeedbackStatus(std::string("Couldn't get parameters for OpenCL program: ") + m_pLens->getErrorString());
		return false;
	}

	std::string program, failLog, subRoutine;

	program = m_pLens->getCLLensProgram(subRoutine);

	program += "\n";
	program += "__kernel void renderLensPlane(float2 startCoord, float2 step, int numX, int numY,\n";
	program += "                     __global const int *pIntParams,\n";
	program += "		     __global const float *pFloatParams,\n";
	program += "                     __global float *pResults,\n";
	program += "			int subStart, int numSub)\n";
	program += "{\n";
	program += "	const int i = get_global_id(0);\n";
	program += "	const int j = get_global_id(1);\n";
	program += "\n";
	program += "	if (i >= numX || j >= numY)\n";
	program += "		return;\n";
	program += "\n";
	program += "	float2 pos = { startCoord.x + step.x*((float)i), startCoord.y + step.y*((float)j) } ;\n";
	program += "\n";
	program += "	LensQuantities r = " + subRoutine + "(pos, pIntParams, pFloatParams, subStart, numSub);\n";
	program += "\n";
	program += "	const int offset = (i+j*numX)*5;\n";
	program += "\n";
	program += "	pResults[offset+0] += r.alphaX;\n";
	program += "	pResults[offset+1] += r.alphaY;\n";
	program += "	pResults[offset+2] += r.axx;\n";
	program += "	pResults[offset+3] += r.ayy;\n";
	program += "	pResults[offset+4] += r.axy;\n";
	program += "}\n";

	//std::cout << program << std::endl;

	if (!pCLKernel->loadKernel(program, "renderLensPlane", failLog))
	{
		setFeedbackStatus(std::string("Couldn't load OpenCL kernel: ") + pCLKernel->getErrorString());
		if (!failLog.empty())
			std::cerr << failLog << std::endl;
		return false;
	}

	// Got kernel, render using OpenCL

	int numtotal = numx*numy;

	std::vector<float> hostAlphaX(numtotal), hostAlphaY(numtotal);
	std::vector<float> hostResults(numtotal*5);
	cl_int err = 0;
	cl_int err1 = 0;
	cl_int err2 = 0;
	cl_int err3 = 0;
	cl_context context = pCLKernel->getContext();

	memset(&(hostResults[0]), 0, sizeof(float)*hostResults.size());

	cl_mem clIntValues = clCreateBuffer(context, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, sizeof(int) * (numIntParams+1), &(intParams[0]), &err1);
	cl_mem clFloatValues = clCreateBuffer(context, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, sizeof(float) * (numFloatParams+1), &(floatParams[0]), &err2);
	cl_mem clResults = clCreateBuffer(context, CL_MEM_WRITE_ONLY|CL_MEM_COPY_HOST_PTR, sizeof(float) * numtotal * 5, &(hostResults[0]), &err3);

	err = err1|err2|err3;

	if (err != CL_SUCCESS)
	{
		if (clIntValues) clReleaseMemObject(clIntValues);
		if (clFloatValues) clReleaseMemObject(clFloatValues);
		if (clReleaseMemObject) clReleaseMemObject(clResults);

		setFeedbackStatus("Error allocating memory objects on GPU");
		return false;
	}

	cl_int numX = numx;
	cl_int numY = numy;
	cl_int subStart = -1;
	cl_int subNum = -1;

	cl_float2 startPos = { (float)(bottomleft.getX()/deflectionScale), (float)(bottomleft.getY()/deflectionScale) };
	cl_float2 step = { (float)(xstep/deflectionScale), (float)(ystep/deflectionScale) };
	cl_kernel kernel = pCLKernel->getKernel();

	err |= clSetKernelArg(kernel, 0, sizeof(cl_float2), (void *)&startPos);
	err |= clSetKernelArg(kernel, 1, sizeof(cl_float2), (void *)&step);
	err |= clSetKernelArg(kernel, 2, sizeof(cl_int), (void *)&numX);
	err |= clSetKernelArg(kernel, 3, sizeof(cl_int), (void *)&numY);
	err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), (void *)&clIntValues);
	err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void *)&clFloatValues);
	err |= clSetKernelArg(kernel, 6, sizeof(cl_mem), (void *)&clResults);
	err |= clSetKernelArg(kernel, 7, sizeof(cl_int), (void *)&subStart);
	err |= clSetKernelArg(kernel, 8, sizeof(cl_int), (void *)&subNum);

	if (err != CL_SUCCESS)
	{
		clReleaseMemObject(clIntValues);
		clReleaseMemObject(clFloatValues);
		clReleaseMemObject(clResults);

		setFeedbackStatus("Error setting kernel arguments");
		return false;
	}

	setFeedbackStatus("Building map using OpenCL");

	cl_command_queue commandQueue = pCLKernel->getCommandQueue();

	if (linePerLine)
	{
		if (numSubLenses < 0)
		{
			for (int y = 0 ; err == CL_SUCCESS && y < numY ; y++)
			{
				size_t lineStart[2] = { 0, y };
				size_t lineSize[2] = { numX, 1 };

				err |= clEnqueueNDRangeKernel(commandQueue, kernel, 2, lineStart, lineSize, NULL, 0, NULL, NULL);
				err |= clFinish(commandQueue);

				int pct = (int)(((((double)y)/((double)(numY-1))))*100.0+0.5);
				setFeedbackPercentage(pct);
			}
		}
		else
		{
			setFeedbackStatus("TODO");
			err = -1;
		}
	}
	else
	{
		if (numSubLenses < 0)
		{
#if 1
			size_t workSize[2] = { numX, numY };
			cl_command_queue commandQueue = pCLKernel->getCommandQueue();

			err |= clEnqueueNDRangeKernel(commandQueue, kernel, 2, NULL, workSize, NULL, 0, NULL, NULL);
			err |= clFinish(commandQueue);
#else
			setFeedbackStatus("TODO: temporarily disabled");
			err = -1;
#endif
		}
		else
		{
			int totalSubLenses = m_pLens->getCLSubLenses();

			if (totalSubLenses < 0)
			{
				setFeedbackStatus("Sublens rendering not available for this type of lens");
				err = -1;
			}
			else
			{
				int count = 0;

				while (count < totalSubLenses && err == CL_SUCCESS)
				{
					//std::cout << count << "/" << totalSubLenses << std::endl;
					cl_int start = count;
					cl_int num = numSubLenses;

					if (start + num > totalSubLenses)
						num = totalSubLenses-start;


					size_t workSize[2] = { numX, numY };
					cl_command_queue commandQueue = pCLKernel->getCommandQueue();

					err |= clSetKernelArg(kernel, 7, sizeof(cl_int), (void *)&start);
					err |= clSetKernelArg(kernel, 8, sizeof(cl_int), (void *)&num);

					err |= clEnqueueNDRangeKernel(commandQueue, kernel, 2, NULL, workSize, NULL, 0, NULL, NULL);
					err |= clFinish(commandQueue);

					count += num;

					int pct = (int)(((((double)count)/((double)(totalSubLenses))))*100.0+0.5);
					setFeedbackPercentage(pct);
				}
			}
		}
	}

	bool retValue = true;

	if (err != CL_SUCCESS)
	{
		renderError = true;
		retValue = false;
	}
	else
	{
		err |= clEnqueueReadBuffer(commandQueue, clResults, true, 0, sizeof(float)*(numtotal*5), &(hostResults[0]), 0, NULL, NULL);
		err |= clFinish(commandQueue);

		int i5 = 0;

		for (int i = 0 ; i < numtotal ; i++, i5 += 5)
		{
			alphas[i] = Vector2D<double>(hostResults[i5+0], hostResults[i5+1]) * deflectionScale;
			alphaxx[i] = hostResults[i5+2];
			alphayy[i] = hostResults[i5+3];
			alphaxy[i] = hostResults[i5+4];
		}
	}

	clReleaseMemObject(clIntValues);
	clReleaseMemObject(clFloatValues);
	clReleaseMemObject(clResults);

	return retValue;
#endif // GRALECONFIG_SUPPORT_OPENCL
}

} // end namespace

