/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "triangulation.h"
#include "Wm5Delaunay2.h"
#include "ttl/halfedge/HeTriang.h"
#include "ttl/halfedge/HeDart.h"
#include "ttl/halfedge/HeTraits.h"
#include <list>
#include <map>

#include "debugnew.h"

namespace grale
{

Triangulation::Triangulation()
{
}

Triangulation::~Triangulation()
{
}

bool Triangulation::create(const std::vector<Vector2D<double> > &points)
{
	int numPoints = points.size();

	if (numPoints <= 2)
	{
		setErrorString("At least three points are needed");
		return false;
	}

	m_triangles.clear();

	std::vector<Wm5::Vector2<double> > inputPoints(numPoints);

	for (int i = 0 ; i < numPoints ; i++)
		inputPoints[i] = Wm5::Vector2<double>(points[i].getX(), points[i].getY());

	Wm5::Delaunay2<double> delaunay(numPoints, &(inputPoints[0]), 0, false, Wm5::Query::QT_INTEGER);

	int numTriangles = delaunay.GetNumSimplices();
	const int *pIndices = delaunay.GetIndices();

	for (int i = 0 ; i < numTriangles ; i++)
		m_triangles.push_back(TriangleIndices(pIndices[i*3+0], pIndices[i*3+1], pIndices[i*3+2]));
	
	return true;
}

class InternalNode : public hed::Node
{
public:
	InternalNode(double x, double y, int id) : hed::Node(x, y)
	{
		m_id = id;
	}

	int getID() const
	{
		return m_id;
	}
private:
	int m_id;
};

bool Triangulation::create(const std::vector<Vector2D<double> > &points, const std::vector<std::pair<int, int> > &constraints)
{
	hed::Triangulation triang;	

	std::vector<hed::Node *> nodes;
	
	for (int i = 0 ; i < points.size() ; i++)
		nodes.push_back(new InternalNode(points[i].getX(), points[i].getY(), i));

	triang.createDelaunay(nodes.begin(), nodes.end());

	std::map<int, hed::Edge *> nodeEdges;
	std::list<hed::Edge *> *pAllEdges = triang.getEdges();
	std::list<hed::Edge *>::const_iterator eIt;

	for (eIt = pAllEdges->begin() ; eIt != pAllEdges->end() ; eIt++)
	{
		hed::Edge *pEdge = *eIt;
		InternalNode *pNode = (InternalNode *)pEdge->getSourceNode();

		nodeEdges[pNode->getID()] = pEdge;
	}

	delete pAllEdges;

	for (int i = 0 ; i < constraints.size() ; i++)
	{
		hed::Dart d1(nodeEdges[constraints[i].first]);
		hed::Dart d2(nodeEdges[constraints[i].second]);

		hed::Dart dNew = ttl::insertConstraint<hed::TTLtraits>(d1, d2, true);
		dNew.getEdge()->setConstrained();
	}

	m_triangles.clear();

	const std::list<hed::Edge *> &triangles = triang.getLeadingEdges();
	std::list<hed::Edge *>::const_iterator it;

	for (it = triangles.begin() ; it != triangles.end() ; it++)
	{
		hed::Edge *pEdge = *it;
		int ids[3];

		ids[0] = ((InternalNode *)pEdge->getSourceNode())->getID();
		ids[1] = ((InternalNode *)pEdge->getTargetNode())->getID();
		ids[2] = ((InternalNode *)pEdge->getNextEdgeInFace()->getTargetNode())->getID();
		m_triangles.push_back(TriangleIndices(ids[0], ids[1], ids[2]));
	}

	/*
	TODO: check for memory leak
	      deleting pNode seems to generate an error
	for (int i = 0 ; i < points.size() ; i++)
	{
		InternalNode *pNode = (InternalNode *)nodes[i];
		delete pNode;
	}
	*/
	setErrorString("Not finished yet");
	return false;


	return true;
}

void Triangulation::getTriangles(std::list<TriangleIndices> &t) const
{
	t = m_triangles;
}

} // end namespace

