/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"

#ifdef GRALECONFIG_SUPPORT_OPENCL

#include "openclkernel.h"
#include <stdio.h>
#include <iostream>

namespace grale
{

OpenCLKernel::OpenCLKernel()
{
	m_init = false;
	m_context = 0;
	m_queue = 0;
	m_program = 0;
	m_kernel = 0;
	m_pDevices = 0;
}

OpenCLKernel::~OpenCLKernel()
{
	destroy();
}

bool OpenCLKernel::init()
{
	if (m_init)
	{
		setErrorString("OpenCL kernel loader is already initialized");
		return false;
	}

	m_currentProgram = "";
	m_currentKernel = "";

	cl_platform_id platforms[16];
	cl_uint numPlatforms, numDevices;
	cl_int err = clGetPlatformIDs(16, platforms, &numPlatforms);

	if (err != CL_SUCCESS)
	{
		setErrorString(getCLErrorString(err));
		releaseAll();
		return false;
	}

	if (numPlatforms < 1)
	{
		setErrorString("No OpenCL platforms available");
		releaseAll();
		return false;
	}

	cl_platform_id platform = platforms[0]; // TODO: make this configurable?

	err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, NULL, &numDevices);
	if (err != CL_SUCCESS)
	{
		setErrorString("Can't find any GPU devices");
		releaseAll();
		return false;
	}

	m_pDevices = new cl_device_id [numDevices];

	clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, numDevices, m_pDevices, NULL);

	m_context = clCreateContext(0, 1, m_pDevices, NULL, NULL, &err);
	if (err != CL_SUCCESS)
	{
		setErrorString(std::string("Can't create OpenCL context: ") + getCLErrorString(err));
		releaseAll();
		return false;
	}

	m_deviceIndex = 0; // TODO: make this configurable?

	m_queue = clCreateCommandQueue(m_context, m_pDevices[m_deviceIndex], 0, &err);
	if (err != CL_SUCCESS)
	{
		setErrorString(std::string("Can't create OpenCL command queue: ") + getCLErrorString(err));
		return false;
	}

	m_init = true;

	return true;
}

bool OpenCLKernel::loadKernel(const std::string &programString, const std::string &kernelName, std::string &failLog)
{
	if (!m_init)
	{
		setErrorString("OpenCL kernel loader was not initialized");
		return false;
	}

	if (programString.length() == 0 || kernelName.length() == 0)
	{
		setErrorString("No program or no kernel name specified");
		return false;
	}

	if (m_program == 0 || (m_program != 0 && programString != m_currentProgram))
	{
		// compile
		
		if (m_program)
			clReleaseProgram(m_program);
		if (m_kernel)
			clReleaseKernel(m_kernel);

		m_program = 0;
		m_kernel = 0;
		m_currentProgram = "";
		m_currentKernel = "";

		const char *pProgStr = programString.c_str();
		cl_int err;

		m_program = clCreateProgramWithSource(m_context, 1, &pProgStr, 0, &err);
		if (err != CL_SUCCESS)
		{
			setErrorString(std::string("Unable to create OpenCL program from specified source: ") + getCLErrorString(err));
			return false;
		}

		err = clBuildProgram(m_program, 1, &(m_pDevices[m_deviceIndex]), "", 0, 0);
		if (err != CL_SUCCESS)
		{
			size_t logLength;

			clGetProgramBuildInfo(m_program, m_pDevices[m_deviceIndex], CL_PROGRAM_BUILD_LOG, 0, 0, &logLength);

			char *pLog = new char[logLength+1];

			clGetProgramBuildInfo(m_program, m_pDevices[m_deviceIndex], CL_PROGRAM_BUILD_LOG, logLength, pLog, 0);
			pLog[logLength] = 0;

			failLog = std::string(pLog);
			delete [] pLog;

			clReleaseProgram(m_program);
			m_program = 0;

			setErrorString(std::string("Unable to build OpenCL program: ") + getCLErrorString(err));

			return false;
		}

		m_currentProgram = programString;
	}

	if (m_kernel == 0 || (m_kernel != 0 && m_currentKernel != kernelName))
	{
		cl_int err;

		if (m_kernel)
			clReleaseKernel(m_kernel);

		m_currentKernel = "";

		m_kernel = clCreateKernel(m_program, kernelName.c_str(), &err);
		if (err != CL_SUCCESS)
		{
			setErrorString(std::string("Unable to get OpenCL kernel from program: ") + getCLErrorString(err));
			return false;
		}

		m_currentKernel = kernelName;
	}

	return true;
}

bool OpenCLKernel::destroy()
{
	if (!m_init)
	{
		setErrorString("OpenCL kernel loader was not initialized");
		return false;
	}

	m_init = false;

	m_currentProgram = "";
	m_currentKernel = "";

	releaseAll();

	return true;
}

void OpenCLKernel::releaseAll()
{
	if (m_kernel)
		clReleaseKernel(m_kernel);
	if (m_program)
		clReleaseProgram(m_program);
	if (m_queue)
		clReleaseCommandQueue(m_queue);
	if (m_context)
		clReleaseContext(m_context);
	if (m_pDevices)
		delete [] m_pDevices;

	m_context = 0;
	m_queue = 0;
	m_program = 0;
	m_kernel = 0;
	m_pDevices = 0;
}

std::string OpenCLKernel::getCLErrorString(int errNum)
{
	char str[256];

	sprintf(str, "OpenCL error code %d", errNum);

	return std::string(str);
}

} // end namespace

#endif // GRALECONFIG_SUPPORT_OPENCL
